/*
 * Copyright (c) 2022 Libre Solar Technologies GmbH
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#include <zephyr/kernel.h>
#include <zephyr/device.h>
#include <zephyr/drivers/uart.h>

#include <string.h>

/* change this to any other UART peripheral if desired */
#define UART_DEVICE_NODE DT_CHOSEN(zephyr_shell_uart)

#define MSG_SIZE  1024

/* queue to store up to 10 messages (aligned to 4-byte boundary) */
K_MSGQ_DEFINE(uart_msgq, MSG_SIZE, 10, 4);

//static const struct device *const uart_dev = DEVICE_DT_GET(UART_DEVICE_NODE);
static const struct device *const uart_dev = DEVICE_DT_GET(DT_NODELABEL(uart1));

const struct uart_config uart_cfg = {
		//.baudrate = 115200,
		.baudrate = 921600,
		.parity = UART_CFG_PARITY_NONE,
		.stop_bits = UART_CFG_STOP_BITS_1,
		.data_bits = UART_CFG_DATA_BITS_8,
		.flow_ctrl = UART_CFG_FLOW_CTRL_NONE
	};

/* receive buffer used in UART ISR callback */
static char rx_buf[MSG_SIZE];
static int rx_buf_pos;

char gUartTxBuffer[MSG_SIZE];
volatile int gUartTxLen;
volatile uint8_t rxFlag;

static volatile bool data_transmitted;
static volatile bool uart_tx_ready = true;
static volatile bool data_received;

/*
 * Read characters from UART until line end is detected. Afterwards push the
 * data to the message queue.
 */
void serial_cb(const struct device *dev, void *user_data)
{
	//uint8_t c[10] = {0};
	uint8_t c = 0 ;

	if (!uart_irq_update(uart_dev)) {
		return;
	}

   if (uart_irq_tx_complete(dev)) {
      uart_tx_ready = true;
   }

   if (uart_irq_tx_ready(dev)) {
      uart_irq_tx_disable(dev);
      uart_fifo_fill(dev, (uint8_t*)&gUartTxBuffer, gUartTxLen);
   }

	if (uart_irq_rx_ready(uart_dev)) {

	/* read until FIFO empty */
   while (uart_fifo_read(uart_dev, &c, 1) == 1) {
		if ((c == '\n' || c == '\r') && rx_buf_pos > 0) {
			/* terminate string */
			rx_buf[rx_buf_pos] = '\0';
         rxFlag = 1;

			/* if queue is full, message is silently dropped */
			k_msgq_put(&uart_msgq, &rx_buf, K_NO_WAIT);

			/* reset the buffer (it was copied to the msgq) */
			rx_buf_pos = 0;
		} else if (rx_buf_pos < (sizeof(rx_buf) - 1)) {
			rx_buf[rx_buf_pos++] = c;
		}
		/* else: characters beyond buffer size are dropped */
	}
	}
}

/*
 * Print a null-terminated string character by character to the UART interface
 */
void print_uart(const struct device *dev, char *buf)
{
   int msg_len = strlen(buf);
   int ret = 0;
   static int buff_len = 0;
   gUartTxLen = strlen(buf);
   printf("str_len : %d\r\n ", msg_len);
   //printf(" buf : %s\n", buf);
#if 0
   buff_len = 512;
   printf("buff_len : %d\n", buff_len);

   printf("str_len : %d\n", strlen(tx_buff));
   memcpy(gUartTxBuffer, tx_buff, MSG_SIZE);
#endif

   memset(gUartTxBuffer, 0x00, MSG_SIZE);
   memcpy(gUartTxBuffer, buf, gUartTxLen);

   if (uart_tx_ready) {
      uart_tx_ready = false;
      uart_irq_tx_enable(uart_dev);
   }
}

int main(void)
{
   char tx_buf[MSG_SIZE];
   int ret = 0;

   char tx_buff_1[1024] = "0123456789-#1#/-0123456789-#2#/-0123456789-#3#/-0123456789-#4#/-0123456789-#5#/-0123456789-#6#/-0123456789-#7#/-0123456789-#8#/-0123456789-#9#/-0123456789-#10#/-0123456789-#11#/-0123456789-#12#/-0123456789-#13#/-0123456789-#14#/-0123456789-#15#/-0123456789-#16#/-0123456789-#17#/-0123456789-#18#/-0123456789-#19#/-0123456789-#20#/-0123456789-#21#/-0123456789-#22#/-0123456789-#23#/-0123456789-#24#/-0123456789-#25#/-0123456789-#26#/-0123456789-#27#/-0123456789-#28#/-0123456789-#29#/-0123456789-#30#/-0123-#31#/-0123456789-#32#/-0123456789-#33#/-0123456789-#34#/-0123456789-#35#/-0123456789-#36#/-0123456789-#37#/-0123456789-#38#/-0123456789-#39#/-0123456789-#40#/-0123456789-#41#/-0123456789-#42#/-0123456789-#43#/-0123456789-#44#/-0123456789-#45#/-0123456789-#46#/-0123456789-#47#/-0123456789-#48#/-0123456789-#49#/-0123456789-#50#/-0123456789-#51#/-0123456789-#52#/-0123456789-#53#/-0123456789-#54#/-0123456789-#55#/-0123456789-#56#/-0123456789-#57#/-0123456789-#58#/-0123456789-#59#/-01234\r\n"

   if (!device_is_ready(uart_dev)) {
      printf("UART device not found!");
      return 0;
   }

   /* configure interrupt and callback to receive data */
   ret = uart_irq_callback_user_data_set(uart_dev, serial_cb, NULL);
   if (ret < 0) {
      if (ret == -ENOTSUP) {
         printf("Interrupt-driven UART API support not enabled\n");
      } else if (ret == -ENOSYS) {
         printf("UART device does not support interrupt-driven API\n");
      } else {
         printf("Error setting UART callback: %d\n", ret);
      }
      return 0;
   }
   uart_irq_rx_enable(uart_dev);

   for (int i = 0; i< 5; i++ ) {
         print_uart(uart_dev, "Hello! I'm your echo bot.\r\n");
         k_msleep(1);
         print_uart(uart_dev, "Tell me something and press enter:\r\n");
   }

   k_msleep(5000);
   print_uart(uart_dev, tx_buff_1);

   /* indefinitely wait for input from the user */
   while (k_msgq_get(&uart_msgq, &tx_buf, K_FOREVER) == 0) {
         printf("received buff: %s\r\n", tx_buf);
   }
   return 0;
}
